#include "cli.h"

static fmus_err_t parse_cli_args_glyph_mode(fmus_t *fmus,
                                            int argc,
                                            char *argv[],
                                            char *v,
                                            int *argix_inout,
                                            fmus_glyph_t *glyph_out,
                                            u8_t *mode_out,
                                            u8_t *gelixes_out);
static fmus_err_t parse_cli_arg_normal_mode(fmus_t *fmus,
                                            int argc,
                                            char *argv[],
                                            char *v,
                                            int *argix_inout,
                                            u32_t *num_unknowns_inout,
                                            u8_t *glyph_mode_out);
static void print_usage (fmus_t *fmus);
static void error_list_gels(u32_t tw);

// see also GELIX_... constants in fmus.h
const char* glyph_edit_labels[] = {
  "txl",
  "txt",
  "txr",
  "txb",
  "lsb",
  "lwidth",
  "rsb",
  "dwidth",
  "height",
  "vertoff"
};

static fmus_err_t parse_cli_args_glyph_mode(fmus_t *fmus,
                                            int argc,
                                            char *argv[],
                                            char *v,
                                            int *argix_inout,
                                            fmus_glyph_t *glyph_out,
                                            u8_t *mode_out,
                                            u8_t *gelixes_out) { // list of which glyph fields were modified
                                              
  u32_t utf32=0;
  //u32_t q;
  u8_t matched=0;
  char *v2;
  int result;
  u32_t i, j;
  char *p;
  size_t v_len;
  
  //*mode_out = GLYPHMOD_NONE;
  
  v = argv[*argix_inout];
  
//printf("v=%s\n", v);
//printf("argix_inout = %u\n", *argix_inout);

  memset(glyph_out, 0, sizeof(fmus_glyph_t));
  
  result = -1;
  if (!strcmp(v, "add") || !strcmp(v, "delete") || !strcmp(v, "edit")) {
    if (((*argix_inout) + 1) >= argc) {
#define CLI_ERRMSG_GLYPH_EDIT_UTF16 "Error: Glyph %s expects a 16-bit Unicode value in hexadecimal. " \
                                    "(You can optionally prefix this with 0x or 0X.)\n"
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GLYPH_EDIT_UTF16, argv[(*argix_inout)]);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    v2 = argv[*argix_inout];
    // expect UTF16 hex

    p = v2;
    v_len = strlen(p);
    if (v_len > 2) {
      if (!strncmp(p, "0x", 2) || !strncmp(p, "0X", 2)) {
        p = v2+2;
        v_len -= 2;
      }
    }
    if (v_len == 0) {
#define CLI_ERRMSG_GLYPH_EDIT_UTF16 "Error: Glyph %s expects a 16-bit Unicode value in hexadecimal. " \
                                    "(You can optionally prefix this with 0x or 0X.)\n"
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GLYPH_EDIT_UTF16, argv[(*argix_inout)]);
      return FMUS_E_CLI;
    }
//printf("lol\n");
    for (j=0;j<v_len;j++) { // sanitise input
      if (((p[j]<'0') || (p[j]>'9')) && ((p[j]<'a') || (p[j]>'f')) && ((p[j]<'A') || (p[j]>'F'))) {
        result = 0;
        break;
      }
    }
//printf("%s\n", p);
    if (result == -1)
      result = FMUS_SSCANF (p, "%x", &utf32);
    if (!result) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GLYPH_EDIT_UTF16, argv[(*argix_inout)-1]);
      return FMUS_E_CLI;
    }
    
    glyph_out->utf16 = 0xffff & utf32;
  
  } else {
    FMUS_EPRINTF(fmus->tw, "Error: %s expects one of \"add\", \"edit\" or \"delete\".\n", argv[(*argix_inout)-1]);
    return FMUS_E_CLI;
  }
  
  if (utf32 > 0xffff) {
    FMUS_EPRINTF(fmus->tw, "Error: Glyph Unicode values greater than 0xffff are not supported.\n");
    return FMUS_E_CLI;
  }
  
  if (strcmp(v, "edit")) { // anything but edit
    glyph_out->utf16 = 0xffff & utf32;
    if (!strcmp(v, "add"))
      *mode_out = GLYPHMOD_ADD; // add
    else
      *mode_out = GLYPHMOD_DELETE; // delete
    return FMUS_E_OK;
  }
  
  *mode_out = GLYPHMOD_EDIT; // edit
  
  // edit mode confirmed, need more args
  
  //do {
  
  if (((*argix_inout) + 1) >= argc) {
    error_list_gels(fmus->tw);
    return FMUS_E_CLI;
  }
  
  matched=0;
  
  memset (gelixes_out, 0, NUM_GLYPH_EDIT_LABELS);

  (*argix_inout)++;
  v = argv[*argix_inout];
  
//printf("v=%s\n", v);
  
  for (i=0;i<NUM_GLYPH_EDIT_LABELS;i++) {
    float f;
    u32_t u;
    u32_t j;
    s32_t s32;
    size_t v_len;
    result = -1;
    if (!strcmp(glyph_edit_labels[i], v)) {
//printf("%s = %s\n", glyph_edit_labels[i], v);
      //(*argix_inout)++;
      if (((*argix_inout) + 1) >= argc) {
        FMUS_EPRINTF(fmus->tw,
                     "Error: Glyph edit mode expects a value after the field label (see documentation).\n");
        return FMUS_E_CLI;
      }
      (*argix_inout)++;
      v = argv[*argix_inout];
//printf("v=%s\n", argv[*argix_inout]);
      v_len = strlen(v);
//printf("v=%s\n", v);
      switch (i) {
        case GELIX_TXCL: case GELIX_TXCT: case GELIX_TXCR: case GELIX_TXCB:
          // texture coordinates are floats
          for (j=0;j<v_len;j++) { // sanitise input
//printf("%c\n", v[j]);
            if (((v[j]<'0') || (v[j]>'9')) && (v[j] != '.')) {
              result = 0;
              break;
            }
          }
          if (result == -1)
            result = FMUS_SSCANF (v, "%f", &f);
          if (!result || (result < 0.0) || (result > 1.0)) {
            FMUS_EPRINTF(fmus->tw,
                         "Error: Glyph texture coordinates must be decimal numbers between 0.0 and 1.0.\n");
            return FMUS_E_CLI;
          }
          break;
        case GELIX_LWIDTH: case GELIX_DWIDTH: case GELIX_HEIGHT: case GELIX_VERTOFF:
          // U8s
          for (j=0;j<v_len;j++) { // sanitise input
            if (((v[j]<'0') || (v[j]>'9'))) {
              result = 0;
              break;
            }
          }
          if (result == -1)
            result = FMUS_SSCANF (v, "%u", &u);
          if (!result || (u > 255)) {
            FMUS_EPRINTF(fmus->tw,
                         "Error: Glyph field %s must be an integer of 255 or less.\n",
                         glyph_edit_labels[i]);
            return FMUS_E_CLI;
          }
          break;
        case GELIX_LSB: case GELIX_RSB:
          // S8s
          for (j=0;j<v_len;j++) { // sanitise input
            if (((v[j]<'0') || (v[j]>'9')) && (v[j] != '-')) {
              result = 0;
              break;
            }
          }
          if (result == -1)
            result = FMUS_SSCANF (v, "%i", &s32);
          if (!result || (s32 < -128) || (s32 > 127)) {
            FMUS_EPRINTF(fmus->tw,
                         "Error: Glyph field %s must be an integer between -128 and 127.\n",
                         glyph_edit_labels[i]);
            return FMUS_E_CLI;
          }
          break;
      }
      switch (i) {
        case GELIX_TXCL:    glyph_out->tx_left = f; break;
        case GELIX_TXCT:    glyph_out->tx_top = f; break;
        case GELIX_TXCR:    glyph_out->tx_right = f; break;
        case GELIX_TXCB:    glyph_out->tx_bottom = f; break;
        case GELIX_LSB:     glyph_out->lsb = s32 & 0xff; break;
        case GELIX_LWIDTH:  glyph_out->layout_width = u & 0xff; break;
        case GELIX_RSB:     glyph_out->rsb = s32 & 0xff; break;
        case GELIX_DWIDTH:  glyph_out->display_width = u & 0xff; break;
        case GELIX_HEIGHT:  glyph_out->height = u & 0xff; break;
        case GELIX_VERTOFF: glyph_out->vert_offset = u & 0xff; break;
        default: break; // fixme: bug
      }
      gelixes_out[i] = 1;
      matched=1;
    }
  } // next GEL to try 
  //} while (matched);
  
  if (!matched) {
    error_list_gels(fmus->tw);
    return FMUS_E_CLI;
  }
  
  return FMUS_E_OK;

}

static void error_list_gels(u32_t tw) {
  u32_t q;
  FMUS_EPRINTF(tw, "Error: Glyph edit operation requires you to specify "
                         "which field of the glyph you wish to change. Possible "
                         "values are:\n");
  for (q=0;q<NUM_GLYPH_EDIT_LABELS;q++) {
    FMUS_EPRINTF(tw, "%s%s", ((q>0) ? ", " : ""), glyph_edit_labels[q]);
  }
  printf("\n");
}

fmus_err_t parse_cli(fmus_t *fmus, int argc, char *argv[]) {
  
  int j;
  //fmus_glyph_t tmpglyph;
  u8_t mode = 0;
  u32_t num_unknowns = 0;
  gclip_err_t e;
  
  e = FMUS_E_OK;
  
  if (argc < 2) {
    print_usage(fmus);
    return FMUS_E_CLI;
  }
  
  for (j=1;j<argc;j++) {
    char *v = argv[j];
    //s32_t result;
    u8_t gelixes_to_change[NUM_GLYPH_EDIT_LABELS];
    fmus_glyph_t glyph;
    if (mode == GLYPHMOD_NONE) { // normal (non -g) mode
//printf("n %s\n", v);
      e = parse_cli_arg_normal_mode (fmus, argc, argv, v, &j, &num_unknowns, &mode);
      if (e != FMUS_E_OK)
        return e;
    } else { // -g mode (>0)
//printf("g %s\n", v);
      e = parse_cli_args_glyph_mode (fmus, argc, argv, v, &j, &glyph, &mode, gelixes_to_change);
      if (e != FMUS_E_OK)
        return e;
      // have glyph and fields that are to be changed on it
      if (fmus->glyphmods_fill >= fmus->glyphmods_alloced) {
        // realloc glyphmods
        if (fmus->glyphmods_alloced == 0)
          fmus->glyphmods_alloced = 1;
        else
          (fmus->glyphmods_alloced) *= 2;
        fmus->glyphmods = FMUS_REALLOCMSG(fmus->glyphmods,
                                          fmus->glyphmods_alloced * sizeof(fmus_glyphmod_t),
                                          "glyphmods");
//printf("memset(%p+%u, 0, %u)\n", fmus->glyphmods, fmus->glyphmods_fill, (fmus->glyphmods_alloced - fmus->glyphmods_fill) * sizeof(fmus_glyphmod_t));
        memset(fmus->glyphmods + fmus->glyphmods_fill,
               0,
               (fmus->glyphmods_alloced - fmus->glyphmods_fill) * sizeof(fmus_glyphmod_t));
      }
//printf("glyphmods[%u].mod_type=%u\n", fmus->glyphmods_fill, mode);
      fmus->glyphmods[fmus->glyphmods_fill].mod_type = mode; // add, edit, delete
      fmus->glyphmods[fmus->glyphmods_fill].glyph = glyph;   // modified data
      memcpy(fmus->glyphmods[fmus->glyphmods_fill].gelixes,  // which fields were modified?
             gelixes_to_change,
             NUM_GLYPH_EDIT_LABELS);
      (fmus->glyphmods_fill)++;
      mode = GLYPHMOD_NONE;
    }
  }
  
//  if (mode != 0) {
    // not allowed to end during a -g mode
//    print_usage(fmus);
//    return FMUS_E_CLI_HELP;
//  }

  if (fmus->query_only && (NULL != fmus->font_filename)) {
    FMUS_EPRINTF(fmus->tw, "Error: You cannot supply an input .FONT filename (%s) "
                           "in query-only mode (-q) or verbose query mode (-Q).\n",
                           fmus->font_filename);
    return FMUS_E_CLI;
  }

  if (fmus->do_not_load && (NULL != fmus->mpfont_filename)) {
    FMUS_EPRINTF(fmus->tw, "Error: You cannot supply an input .FONT filename (%s) in blank-font mode (-b).\n",
                 fmus->mpfont_filename);
    return FMUS_E_CLI;
  }
  
  if (fmus->edit_mode && (NULL != fmus->font_filename)) {
    FMUS_EPRINTF(fmus->tw, "Error: You cannot supply a TTF/OTF filename (%s) in font edit mode (-e).\n",
                 fmus->font_filename);
    return FMUS_E_CLI;
  }

  if (fmus->omit_kerning && !fmus->edit_mode) {
    FMUS_EPRINTF(fmus->tw, "Error: --omit-kerning can only be used in edit mode (-e).\n");
    return FMUS_E_CLI;
  }

  if ((!fmus->edit_mode || fmus->do_not_load) && fmus->no_compact_kerning) {
    FMUS_EPRINTF(fmus->tw, "Error: --no-compact-kerning can only be used in edit mode (-e).\n");
    return FMUS_E_CLI;
  }
  
  if (fmus->query_only && fmus->edit_mode) {
    FMUS_EPRINTF(fmus->tw, "Error: -e and -b cannot be used with -q or -Q.\n");
    return FMUS_E_CLI;
  }
  
  if (!fmus->do_not_load && (NULL == fmus->mpfont_filename)) {
    // no -b: do not allow blank mpfont filename
    print_usage(fmus);
    return FMUS_E_CLI_HELP;
  }
  if (!fmus->query_only && !fmus->edit_mode && !fmus->do_not_load) {
    // no -q, -e, -b: do not allow blank vector font filename
    if (NULL == fmus->font_filename) {
      print_usage(fmus);
      return FMUS_E_CLI_HELP;
    }
  }
  
  if ((fmus->font_filename != NULL) && (fmus->mpfont_filename != NULL)) {
    if (!strcmp(fmus->mpfont_filename, fmus->font_filename)) {
      FMUS_EPRINTF(fmus->tw, "Error: You cannot provide the same filename for both the TTF/OTF "
                             "and the input .FONT file!\n");
      return FMUS_E_CLI;
    }
  }
  
  // blank mode implies edit mode
  if (fmus->do_not_load)
    fmus->edit_mode = 1;
  
  return FMUS_E_OK;
  
}

static void print_usage (fmus_t *fmus) {
  printf("\n");
  FMUS_MPRINTF(fmus->tw, "%s\n", FMUS_VERSION_FULL);
  FMUS_MPRINTF(fmus->tw, "A font manipulation utility for the Metroid Prime series of games.\n");
  FMUS_MPRINTF(fmus->tw, "This software comes with no warranty.\n\n");
  FMUS_MPRINTF(fmus->tw, "Usage:    fp [options] <input .FONT file> <input .TTF/.OTF file>\n");
  FMUS_MPRINTF(fmus->tw, "       or fp -q, -Q, -e [options] <input .FONT file>\n");
  FMUS_MPRINTF(fmus->tw, "       or fp -b [options]\n\n");
  //FMUS_MPRINTF(fmus->tw, "Options:\n");
  FMUS_MPRINTF(fmus->tw,
               "  -q, --query-only                   query input .FONT, perform no action\n"
               "  -Q, --verbose-query                same as -q, but more verbose\n"
               "  -e, --edit-mode                    alter .FONT, do not render to TXTR\n"
               "  -b, --blank-font                   like -e, but start with an empty .FONT\n"
               "  -f, --output-font    <.FONT file>  set output .FONT filename\n"
               "  -t, --texture        <.TXTR file>  set output .TXTR filename\n"
               "  -w, --overwrite                    overwrite existing output files\n"
               "  -s, --text-size      <text size>   set text size (decimal)\n"
               "  -o, --output-font-version          set .FONT version (2, 4 or 5)\n"
               "  -x, --texture-format               set texture format (0-4, or 8)\n"
               "  -n, --font-name      <.FONT name>  set .FONT name\n"
               "  -i, --texture-id     <texture ID>  set .FONT texture ID (hex, 32 or 64-bit)\n"
               "  --line-spacing       <pixels>      set line spacing\n"
               "  -y, --y-scale        <multiplier>  multiply glyph heights by a decimal\n"
               "  --letter-gap-adjust  <+/- pixels>  adjust letter gap\n"
               "  --space-adjust       <+/- pixels>  adjust word space gap\n"
               "  -c, --input-face     <face index>  select input TTF/OTF face (default 0)\n"
               "  -g, --glyph          [options]     edit glyphs -- see the documentation\n"
               "  --no-compact-kerning               with -e, do not compact kerning table\n"
               "  --omit-kerning                     with -e, omit kerning from output .FONT\n"
               "  -v, --version                      print Fontimus Prime version\n"
               "  -h, --help                         show this help text\n"
               );
}

#define FONT_PX_SIZE_MIN 4.0
#define FONT_PX_SIZE_MAX 128.0
#define LGAP_ADJUST_MIN  -64
#define LGAP_ADJUST_MAX  64
#define WGAP_ADJUST_MIN  -64
#define WGAP_ADJUST_MAX  64
#define Y_STRETCH_MIN    0.1
#define Y_STRETCH_MAX    3.0

static fmus_err_t parse_cli_arg_normal_mode(fmus_t *fmus,
                                            int argc,
                                            char *argv[],
                                            char *v,
                                            int *argix_inout,
                                            u32_t *num_unknowns_inout,
                                            u8_t *glyph_mode_out) {
  
  int result;
  u32_t u32;
  s32_t s32;
  
  *glyph_mode_out = GLYPHMOD_NONE;

  if (!strcmp(v, "-f") || !strcmp(v, "--output-font")) {
    if (((*argix_inout) + 1) >= argc) {
      FMUS_EPRINTF(fmus->tw, "Error: %s argument must be followed by an output .FONT file name.\n", v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    fmus->output_font_filename = argv[*argix_inout];
  } else if (!strcmp(v, "-t") || !strcmp(v, "--texture")) {
    if (((*argix_inout) + 1) >= argc) {
      FMUS_EPRINTF(fmus->tw, "Error: %s argument must be followed by an output .TXTR file name.\n", v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    fmus->texture_filename = argv[*argix_inout];
  } else if (!strcmp(v, "-s") || !strcmp(v, "--text-size")) {
    if (((*argix_inout) + 1) >= argc) {
#define CLI_ERRMSG_PX_SIZE "Error: %s argument must be followed by a decimal pixel size " \
                         "between %f and %f.\n"
      FMUS_EPRINTF(fmus->tw,
                   CLI_ERRMSG_PX_SIZE,
                   v,
                   (double) FONT_PX_SIZE_MIN,
                   (double) FONT_PX_SIZE_MAX);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[(*argix_inout)], "%f", &fmus->font_px_size);
    if (result < 1) {
      FMUS_EPRINTF(fmus->tw,
                   CLI_ERRMSG_PX_SIZE,
                   v,
                   (double) FONT_PX_SIZE_MIN,
                   (double) FONT_PX_SIZE_MAX);
      return FMUS_E_CLI;
    }
    if ((fmus->font_px_size < FONT_PX_SIZE_MIN) || (fmus->font_px_size > FONT_PX_SIZE_MAX)) {
      FMUS_EPRINTF(fmus->tw,
                   CLI_ERRMSG_PX_SIZE,
                   v,
                   (double) FONT_PX_SIZE_MIN,
                   (double) FONT_PX_SIZE_MAX);
      return FMUS_E_CLI;
    }
  } else if (!strcmp(v, "-w") || !strcmp(v, "--overwrite")) {
    fmus->overwrite = 1;
  } else if (!strcmp(v, "-o") || !strcmp(v, "--output-font-version")) {
    if (((*argix_inout) + 1) >= argc) {
#define CLI_ERRMSG_FONTVER "Error: %s argument must be followed by an integer font version, " \
                         "either 2, 4 or 5.\n"
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_FONTVER, v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[*argix_inout], "%u", &fmus->op_font_ver);
    if (result < 1) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_FONTVER, v);
      return FMUS_E_CLI;
    }
    if ((fmus->op_font_ver != 2) && (fmus->op_font_ver != 4) && (fmus->op_font_ver != 5)) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_FONTVER, v);
      return FMUS_E_CLI;
    }
    fmus->override_op_font_ver = 1;
  } else if (!strcmp(v, "-x") || !strcmp(v, "--texture-format")) {
    if (((*argix_inout) + 1) >= argc) {
#define CLI_ERRMSG_TXFMT "Error: %s argument must be followed by an integer texture format -- " \
                         "either 0, 1, 2, 3, 4, or 8.\n"
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_TXFMT, v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[*argix_inout], "%u", &fmus->op_tx_fmt);
    if (result < 1) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_TXFMT, v);
      return FMUS_E_CLI;
    }
    if ((fmus->op_tx_fmt > 4) && (fmus->op_tx_fmt != 8)) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_TXFMT, v);
      return FMUS_E_CLI;
    }
    fmus->override_op_tx_fmt = 1;
  } else if (!strcmp(v, "-q")
             || !strcmp(v, "--query-only")) {
    if (fmus->query_only) {
      FMUS_EPRINTF(fmus->tw, "Error: Query mode (-q) cannot be specified "
                             "more than once, or used along with verbose "
                             "query mode (-Q).\n");
      return FMUS_E_CLI;
    }
    if (fmus->edit_mode) {
      FMUS_EPRINTF(fmus->tw, "Error: Query mode (-q) cannot be used along "
                             "with edit mode (-e) or blank mode (-b).\n");
      return FMUS_E_CLI;
    }
    fmus->query_only = 1;
  } else if (!strcmp(v, "-Q")
             || !strcmp(v, "--verbose-query")) {
    if (fmus->query_only) {
      FMUS_EPRINTF(fmus->tw, "Error: Query mode (-q) cannot be specified "
                             "more than once, or used along with verbose "
                             "query mode (-Q).\n");
      return FMUS_E_CLI;
    }
    if (fmus->edit_mode) {
      FMUS_EPRINTF(fmus->tw, "Error: Verbose query mode (-Q) cannot be used along "
                             "with edit mode (-e) or blank mode (-b).\n");
      return FMUS_E_CLI;
    }
    fmus->query_only = 1;
    fmus->verbose_query = 1;
  } else if (!strcmp(v, "-e") || !strcmp(v, "--edit-mode")) {
    if (fmus->query_only) {
      FMUS_EPRINTF(fmus->tw, "Error: Edit mode (-e) cannot be used along with "
                             "query mode (-q) or verbose query mode (-Q).\n");
      return FMUS_E_CLI;
    }
    if (fmus->edit_mode) {
      FMUS_EPRINTF(fmus->tw, "Error: Edit mode (-e) cannot be specified more than once, "
                             "or used along with blank mode (-b).\n");
      return FMUS_E_CLI;
    }
    fmus->edit_mode = 1;
  } else if (!strcmp(v, "-b") || !strcmp(v, "--blank-font")) {
    if (fmus->query_only) {
      FMUS_EPRINTF(fmus->tw, "Error: Blank mode (-b) cannot be used along with "
                             "query mode (-q) or verbose query mode (-Q).\n");
      return FMUS_E_CLI;
    }
    if (fmus->edit_mode) {
      FMUS_EPRINTF(fmus->tw, "Error: Blank mode (-b) cannot be specified more than once, "
                             "or used along with edit mode (-e).\n");
      return FMUS_E_CLI;
    }
    fmus->do_not_load=1;
    fmus->edit_mode=1;
  } else if (!strcmp(v, "-v") || !strcmp(v, "--version")) {
    FMUS_MPRINTF(fmus->tw, "%s\n", FMUS_VERSION_FULL);
    return FMUS_E_CLI_HELP;
  } else if (!strcmp(v, "-h") || !strcmp(v, "--help")) {
    print_usage(fmus);
    return FMUS_E_CLI_HELP;
  } else if (!strcmp(v, "-c") || !strcmp(v, "--input-face")) {
    if (((*argix_inout) + 1) >= argc) {
#define CLI_ERRMSG_FACE "Error: %s argument must be followed by an integer font face index.\n"
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_FACE, v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[*argix_inout], "%u", &fmus->font_face_ix);
    if (result < 1) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_FACE, v);
      return FMUS_E_CLI;
    }
  } else if (!strcmp(v, "-n") || !strcmp(v, "--font-name")) {
    // override font name
    if (((*argix_inout) + 1) >= argc) {
      FMUS_EPRINTF(fmus->tw, "Error: %s argument must be followed by a name string.\n", v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    fmus->override_op_font_name = argv[*argix_inout];
  } else if (!strcmp(v, "-g") || !strcmp(v, "--glyph")) {
    *glyph_mode_out = GLYPHMOD_G;
  } else if (!strcmp(v, "--no-compact-kerning")) {
    fmus->no_compact_kerning = 1;
  } else if (!strcmp(v, "--line-spacing")) {
    if (((*argix_inout) + 1) >= argc) {
#define CLI_ERRMSG_GENERIC "Error: %s argument must be followed by an integer " \
                           "between 0 and 255.\n"
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GENERIC, v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[*argix_inout], "%u", &u32); //&fmus->line_spacing);
    if ((result < 1) || (u32 > 255)) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GENERIC, v);
      return FMUS_E_CLI;
    }
    fmus->override_line_spacing = 1;
    fmus->line_spacing = u32;
  } else if (!strcmp(v, "--y-scale") || !strcmp(v, "-y")) {
#define CLI_ERRMSG_Y_STRETCH "Error: %s argument must be followed by a decimal number " \
                             "between %f and %f.\n"
    if (((*argix_inout) + 1) >= argc) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_Y_STRETCH, v, Y_STRETCH_MIN, Y_STRETCH_MAX);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[*argix_inout], "%f", &fmus->y_stretch);
    if ((result < 1)
        || (fmus->y_stretch > Y_STRETCH_MAX)
        || (fmus->y_stretch < Y_STRETCH_MIN)) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_Y_STRETCH, v, Y_STRETCH_MIN, Y_STRETCH_MAX);
      return FMUS_E_CLI;
    }
    fmus->do_y_stretch = 1;
  } else if (!strcmp(v, "--letter-gap-adjust")) {
#define CLI_ERRMSG_GAPS "Error: %s argument must be followed by an integer " \
                        "between %i and %i.\n"
    if (((*argix_inout) + 1) >= argc) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GAPS, v, LGAP_ADJUST_MIN, LGAP_ADJUST_MAX);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[*argix_inout], "%i", &s32); //&fmus->lgap_adjust);
    if ((result < 1) || (s32 < LGAP_ADJUST_MIN) || (s32 > LGAP_ADJUST_MAX)) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GAPS, v, LGAP_ADJUST_MIN, LGAP_ADJUST_MAX);
      return FMUS_E_CLI;
    }
    fmus->lgap_adjust = s32;
    fmus->do_lgap_adjust = 1;
  } else if (!strcmp(v, "--space-adjust")) {
#define CLI_ERRMSG_GAPS "Error: %s argument must be followed by an integer " \
                        "between %i and %i.\n"
    if (((*argix_inout) + 1) >= argc) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GAPS, v, WGAP_ADJUST_MIN, WGAP_ADJUST_MAX);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    result = FMUS_SSCANF(argv[*argix_inout], "%i", &s32);
    if ((result < 1) || (s32 < WGAP_ADJUST_MIN) || (s32 > WGAP_ADJUST_MAX)) {
      FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_GAPS, v, WGAP_ADJUST_MIN, WGAP_ADJUST_MAX);
      return FMUS_E_CLI;
    }
    fmus->wgap_adjust = s32;
    fmus->do_wgap_adjust = 1;
  } else if (!strcmp(v, "--omit-kerning")) {
    fmus->omit_kerning = 1;
  } else if (!strcmp(v, "-i") || !strcmp(v, "--texture-id")) {
    fmus->txid_string = v; // we can't validate this because we don't know the length yet
#define CLI_ERRMSG_GAPS "Error: %s argument must be followed by an integer " \
                        "between %i and %i.\n"
    if (((*argix_inout) + 1) >= argc) {
      FMUS_EPRINTF(fmus->tw, "Error: %s option must be followed by a hexadecimal "
                             "32 or 64-bit texture ID. It may optionally be prefixed "
                             "with 0x or 0X.\n", v);
      return FMUS_E_CLI;
    }
    (*argix_inout)++;
    fmus->txid_string = argv[*argix_inout]; // argv memory, guaranteed to stick around
  } else {
    // assume input file name
    if (*num_unknowns_inout == 0) // mp font first
      fmus->mpfont_filename = v;
    else if (*num_unknowns_inout == 1) // then vector font
      fmus->font_filename = v;
    else {
      FMUS_EPRINTF(fmus->tw,
                   "Error: unrecognised surplus argument at end of command line: %s\n", 
                   v);
      return FMUS_E_CLI;
    }
    (*num_unknowns_inout)++;
  }
  return FMUS_E_OK;
}

// called later by apply_pre_render_transformations() since we can't validate -i straight away
fmus_err_t cli_parse_txid (fmus_t *fmus, u32_t version, char *txid_s, u64_t *txid_64) {
  size_t txid_len;
  u32_t expected_len;
  u32_t txid_32;
  int result;
  txid_len = strlen(txid_s);
  if ((txid_s[0] == '0') && ((txid_s[1] == 'x') || (txid_s[1] == 'X'))) {
    txid_len -= 2;
    txid_s += 2;
  }
  expected_len = 8;
  if (version == 5)
    expected_len += 8;
  if (txid_len > expected_len) {
#define CLI_ERRMSG_TXID "Error: Supplied texture ID (%s) was not a valid %u-bit hexadecimal number.\n"
    FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_TXID, txid_s, expected_len * 4);
    return FMUS_E_CLI;
  }
  if (version < 5) {
    result = FMUS_SSCANF(txid_s, "%x", &txid_32);
    *txid_64 = txid_32;
  } else
    result = FMUS_SSCANF(txid_s, "%" FMUS_FMT_U64HEX, txid_64);
  if (result == 0) {
    FMUS_EPRINTF(fmus->tw, CLI_ERRMSG_TXID, txid_s, expected_len * 4);
    return FMUS_E_CLI;
  }
  return FMUS_E_OK;
}

