#ifndef FMUS_H
#define FMUS_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// for varargs (slice_printf())
#include <stdarg.h>

#include "fmustypes.h"

typedef s32_t gclip_err_t;

#ifdef FMUS_BIG_ENDIAN
#define FROM_BIG_ENDIAN_64(X) (X)
#define FROM_BIG_ENDIAN_32(X) (X)
#define FROM_BIG_ENDIAN_16(X) (X)
#else
#define FROM_BIG_ENDIAN_64(X) ((((X) << 56) & 0xff00000000000000) \
                             | (((X) << 40) & 0xff000000000000) \
                             | (((X) << 24) & 0xff0000000000) \
                             | (((X) << 8)  & 0xff00000000) \
                             | (((X) >> 8)  & 0xff000000) \
                             | (((X) >> 24) & 0xff0000)) \
                             | (((X) >> 40) & 0xff00) \
                             | (((X) >> 56) & 0xff)
#define FROM_BIG_ENDIAN_32(X) ((((X) << 24) & 0xff000000) \
                             | (((X) << 8)  & 0xff0000) \
                             | (((X) >> 8)  & 0xff00) \
                             | (((X) >> 24) & 0xff))
#define FROM_BIG_ENDIAN_16(X) ((((X) << 8)  & 0xff00) \
                             | (((X) >> 8)  & 0xff))
#endif

#ifndef FMUS_WINDOZE      /* for architectures with ioctl */
#include <sys/ioctl.h>    /* for getting terminal size */
#endif

// for stat():
#ifndef FMUS_WINDOZE
#include <unistd.h>
#include <sys/stat.h>
#endif

#ifdef FMUS_WINDOZE
// for _mkdir()
#include <direct.h>
#endif

// for errors:
#include <errno.h>
#include <ctype.h>

#define FMUS_VERSION_S      "0.1"
#define FMUS_VERSION_I      0x00000001
#define FMUS_VERSION_SHORT  "Fontimus Prime v" FMUS_VERSION_S
#define FMUS_VERSION_TINY   "FP " FMUS_VERSION_S
#define FMUS_VERSION_FULL   FMUS_VERSION_SHORT ", (c) Ali Campbell 2016"

extern char fmus_scprintf_dummy;

#ifdef FMUS_DBGMEM
s32_t fmus_dbgmem_get_allocs(void);
s32_t fmus_dbgmem_get_frees(void);
#endif

#ifdef FMUS_WINDOZE
#define FMUS_INLINE __inline // FIXME
#else
#define FMUS_INLINE inline
#endif

#define WATERMARK_1 " (FPRM "
#define WATERMARK_2 FMUS_VERSION_S ")"
#define WATERMARK   WATERMARK_1 WATERMARK_2

#ifdef FMUS_WINDOZE
#define FMUS_SSCANF                   sscanf_s
#else
#define FMUS_SSCANF                   sscanf
#endif

gclip_err_t gclip_slice_printf(u16_t cols,          /* wrap width */
                               size_t *cursor_pos,  /* current cursor column */
                               size_t wrap_indent,  /* how much to indent new lines */
                               char *fmt,           /* fmtstg */
                               ...);               /* etc. */


typedef struct fmus_glyph {
  u16_t utf16;
  float tx_left;
  float tx_top;
  float tx_right;
  float tx_bottom;
  u32_t rgba_chnl_ix; // added in v4
  s32_t lsb;         // 8-bit in v4
  u32_t layout_width; // 8-bit in v4 (formerly "advance")
  s32_t rsb;    // 8-bit in v4. Definitely signed! 0-00's 192bb274.FONT contains a negative right bearing for 'r'!
  u32_t display_width; // 8-bit in v4
  u32_t height;       // 8-bit in v4
  u32_t vert_offset;  // 8-bit in v4
  u32_t kern_ix;      // "no kerning entry" is represented by kern_ix = kerning_count
} fmus_glyph_t;

typedef struct fmus_kern {
  u16_t c1_utf16;
  u16_t c2_utf16;
  s32_t adjust;
} fmus_kern_t;

#include "mpfont.h"

typedef struct fmus_text fmus_text_t;

// see also glyph_edit_labels in cli.c:
#define GELIX_TXCL    0
#define GELIX_TXCT    1
#define GELIX_TXCR    2
#define GELIX_TXCB    3
#define GELIX_LSB     4
#define GELIX_LWIDTH  5
#define GELIX_RSB     6
#define GELIX_DWIDTH  7
#define GELIX_HEIGHT  8
#define GELIX_VERTOFF 9
#define NUM_GLYPH_EDIT_LABELS 10

#define GLYPHMOD_NONE   0 // normal CLI parsing
#define GLYPHMOD_G      1 // glyph mode (-g)
#define GLYPHMOD_ADD    2 // -g add (also used in glyphmod as the mod_type) 
#define GLYPHMOD_EDIT   3 // -g edit (also used in glyphmod as the mod_type) 
#define GLYPHMOD_DELETE 4 // -g delete (also used in glyphmod as the mod_type) 

typedef struct fmus_glyphmod {
  u8_t mod_type;      // add, edit, delete
  u8_t gelixes[NUM_GLYPH_EDIT_LABELS];       // list of glyph fields to modify (maybe)
  fmus_glyph_t glyph; // contains the modified data (maybe)
} fmus_glyphmod_t;

// an instance
typedef struct fmus_ctx {
  
  u16_t tw;              // terminal width
  fmus_mpfont_t mpfont;  // in-memory font metadata
  u32_t mpfont_glyphtable_start;
  fmus_glyph_t *glyphs;  // in-memory glyph metadata, malloced
  u32_t num_bitmaps;     // num_glyphs * num_layers
  u8_t **bitmaps;        // array of pointers to generated bitmaps. malloced
  fmus_text_t *text;     // freetype stuff
  
  fmus_kern_t *kerning;  // malloced
  
  s32_t ddpi;
  s32_t hdpi;
  s32_t vdpi;
  
  char *mpfont_filename;
  char *font_filename;
  char *output_font_filename;
  char *texture_filename;
  u8_t override_op_font_ver;
  u32_t op_font_ver;
  u8_t override_op_tx_fmt;
  u32_t op_tx_fmt;
  char *override_op_font_name;
  float font_px_size; // -s
  u8_t overwrite;     // -w
  u8_t query_only;    // -q or -Q
  u8_t verbose_query; // -Q
  s32_t font_face_ix;
  u8_t edit_mode;     // -e
  u8_t do_not_load;   // -b
  fmus_glyphmod_t *glyphmods;
  size_t glyphmods_fill;
  size_t glyphmods_alloced;  
  u8_t no_compact_kerning; // edit mode only
  u8_t skip_kerning;       // render mode only
  
  u8_t override_line_spacing;
  u8_t line_spacing;
  u8_t do_y_stretch;
  float y_stretch;
  u8_t do_lgap_adjust; // flag
  u8_t lgap_adjust;    // value
  u8_t do_wgap_adjust;   // flag
  u8_t wgap_adjust;      // value
  
  u8_t omit_kerning; // edit mode only
  char *txid_string; // an argv
  
} fmus_t;

#ifdef FMUS_WINDOZE
float roundf(float f);
char *fmus_windoze_strerror(void);
#endif

#ifdef FMUS_WINDOZE
#define FMUS_STRERROR() fmus_windoze_strerror()
#else
#define FMUS_STRERROR() strerror(errno)
#endif

#ifdef FMUS_WINDOZE
#define FMUS_ISNAN _isnan
#else
#define FMUS_ISNAN isnan
#endif

#endif

