#ifndef FMUSTYPES_H
#define FMUSTYPES_H

// for heap debugging on win32
/*
#ifdef FMUS_WINDOZE
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
#endif
*/

#include <stdio.h>
#include <stdlib.h>

#ifdef _WIN32
#define FMUS_WINDOZE
#include <Windows.h>
#include <sys/types.h>

#ifndef FMUS_NOIOCTL
#define FMUS_NOIOCTL // no ioctl() for terminals on win32
#endif // ndef FMUS_NOIOCTL
#endif // def _WIN32

#define FMUS_E_OK                    0
#define FMUS_E_NULL                  1  // generic NULL pointer on function call
#define FMUS_E_CLI                   2  // bad command line
#define FMUS_E_CLI_HELP              3  // help displayed
#define FMUS_E_MALLOC                4  // malloc() failure
#define FMUS_E_BUG                   5  // generic bug
#define FMUS_E_MALLOC_0              6  // bug: malloc(0)
#define FMUS_E_MKDIR                 7 // could not create dir
#define FMUS_E_FREAD                 8 // fread() errors
#define FMUS_E_FREAD_EOF             9 // input file unexpectedly truncated
#define FMUS_E_FILE_IN_PATH          10 // mkdir: path fragment was a file, not a directory
#define FMUS_E_FWRITE                11 // fwrite() errors
#define FMUS_E_FWRITE_EOF            12 // output file unexpectedly truncated
#define FMUS_E_FOPEN_OPFILE          13 // fopen(output file) failed
#define FMUS_E_FILE_NOT_FILE         14 // could not read or write file as it was a dir or symlink
#define FMUS_E_FILE_EXISTS           15 // tried to overwrite existing file
#define FMUS_E_GETPATH_ZL            16 // fmus_getpath called with zero length string
#define FMUS_E_GETPATH_NOSEP         17 // fmus_getpath no slashes => not a path, just a filename
#define FMUS_E_STAT                  18 // stat() [posix] or GetFileAttributes() [win32] failure
#define FMUS_E_FOPEN_READ_FILE       19 // fopen() in fmus_read_file() failed
#define FMUS_E_ALLOCS_FREES          20 // with FMUS_DBGMEM, exits with error if allocs != frees
#define FMUS_E_FOPEN                 21
#define FMUS_E_FT_LOAD_GLYPH               100
#define FMUS_E_FT_RENDER_GLYPH             101
#define FMUS_E_FT_SET_CHAR_SIZE            102
#define FMUS_E_FT_NEW_FACE                 103
#define FMUS_E_FT_RENDERED_CHAR_TOO_WIDE   104
#define FMUS_E_FT_RENDERED_CHAR_BASELINE_Y 105
#define FMUS_E_FT_RENDERED_CHAR_TOO_TALL   106
#define FMUS_E_FT_GET_CHAR_IX              107
#define FMUS_E_FT_NO_GLYPH_SOLUTION        108
#define FMUS_E_FT_RENDERED_CHAR_L_BEARING  119
#define FMUS_E_FT_FACE_NOT_FOUND           110
#define FMUS_E_FT_MAX_SHEET_SIZE           111
#define FMUS_E_MPF_TRUNC              200
#define FMUS_E_MPF_MAGIC              201
#define FMUS_E_MPF_VERSION            202
#define FMUS_E_MPF_TX_FMT             203
#define FMUS_E_MPF_FONT_SIZE          204
#define FMUS_E_MPF_LINE_MARGIN        205
#define FMUS_E_MPF_VERT_OFFSET        206
#define FMUS_E_MPF_LINE_HEIGHT        207
#define FMUS_E_MPF_NUM_GLYPHS         208
#define FMUS_E_MPF_TX_COORD           209
#define FMUS_E_MPF_LSB                210
#define FMUS_E_MPF_RSB                211
#define FMUS_E_MPF_LAYOUT_WIDTH       212
#define FMUS_E_MPF_KERN_IX            213
#define FMUS_E_MPF_WIDTH              214
#define FMUS_E_MPF_HEIGHT             215
#define FMUS_E_MPF_GLYPH_VERT_OFFSET  216
#define FMUS_E_MPF_KERNLEN            217

#ifndef XYZZY_TYPES
#define XYZZY_TYPES
#ifdef FMUS_WINDOZE
typedef UINT8 u8_t;
typedef INT8 s8_t;
typedef UINT16 u16_t;
typedef INT16 s16_t;
typedef UINT32 u32_t;
typedef INT32 s32_t;
#else // ndef FMUS_WINDOZE
typedef unsigned char u8_t;
typedef signed char s8_t;
typedef unsigned short u16_t;
typedef signed short s16_t;
typedef unsigned int u32_t;
typedef signed int s32_t;
#endif // ndef FMUS_WINDOZE
#endif // ndef XYZZY_TYPES

#ifdef FMUS_WINDOZE
#define FMUS_FUNC_M                            __FUNCTION__
#define FMUS_SNPRINTF(A,B,C,D,...)             _snprintf_s((A),(B),(C),(D),__VA_ARGS__)
//#define FMUS_SCPRINTF(A,...)                   _scprintf((A),__VA_ARGS__)
#define FMUS_STRNCPY(A,B,C,D)                  _mbsncpy_s((A), (B), (C), (D))
#define FMUS_SCPRINTF_1ARG(A,B)       _scprintf((A), (B))
#define FMUS_SSCANF                   sscanf_s
#define GCLIP_SNPRINTF_1ARG(A,B,C,D,E) _snprintf_s((A),(B),(C),(D),(E))
#else
#define FMUS_FUNC_M                            __func__
#define FMUS_SNPRINTF(A,B,C,D...)              snprintf((A),(C),D)
//#define FMUS_SCPRINTF(A...)                    snprintf(&fmus_scprintf_dummy, 0, A)
#define FMUS_STRNCPY(A,B,C,D)                  strncpy((A), (C), (D))
#define FMUS_SCPRINTF_1ARG(A,B)       snprintf(&fmus_scprintf_dummy, 0, (A), (B))
#define FMUS_SSCANF                   sscanf
#define GCLIP_SNPRINTF_1ARG(A,B,C,D,E) snprintf((A),(C),(D),(E))
#endif /* end ndef FMUS_WINDOZE */


#ifdef FMUS_WINDOZE
#define FMUS_UOFF_T      u64_t
#define FMUS_UOFF_T_CAST u64_t
#define FMUS_FMT_UOFF_T  "I64u"
#define FMUS_OFF_T       s64_t
#define FMUS_FMT_OFF_T   "I64d"
#else // ndef FMUS_WINDOZE
#define FMUS_UOFF_T      off_t
#define FMUS_UOFF_T_CAST long long unsigned int
#define FMUS_FMT_UOFF_T  "llu"
#define FMUS_OFF_T       off_t
#define FMUS_FMT_OFF_T   "lld"
#endif

//#ifdef FMUS_64
#ifdef FMUS_WINDOZE
// NB: these are actually the same as the non-windoze
// versions below but in a separate section in case
// I need to change them later for whatever reason
#ifndef FMUS_FMT_U64
typedef UINT64 u64_t;
typedef INT64 s64_t;
#define FMUS_FMT_U64    "I64u"
#endif
#ifndef FMUS_FMT_U64HEX
#define FMUS_FMT_U64HEX "I64x"
#endif
#ifndef FMUS_FMT_S64
#define FMUS_FMT_S64    "I64d"
#endif
#ifndef FMUS_FMT_S64HEX
#define FMUS_FMT_S64HEX "I64x"
#endif
#else // ndef FMUS_WINDOZE
typedef unsigned long long u64_t;
typedef long long s64_t;
#ifndef FMUS_FMT_U64
#define FMUS_FMT_U64    "llu"
#endif
#ifndef FMUS_FMT_U64HEX
#define FMUS_FMT_U64HEX "llx"
#endif
#ifndef FMUS_FMT_S64
#define FMUS_FMT_S64    "lld"
#endif
#ifndef FMUS_FMT_S64HEX
#define FMUS_FMT_S64HEX "llx"
#endif
#endif // ndef FMUS_WINDOZE
//#endif // def FMUS_64

#ifndef FMUS_FMT_SIZET
#ifdef FMUS_WINDOZE
#define FMUS_FMT_SIZET   "Iu"
#define FMUS_FMT_SIZET_X "Ix"
#else
#define FMUS_FMT_SIZET   "zu"
#define FMUS_FMT_SIZET_X "zx"
#endif
#endif

// FIXME: look properly into win32 MSYS pathsep class of bugs
#ifndef FMUS_PATHSEP
#ifdef FMUS_WINDOZE
#define FMUS_PATHSEP '\\'
#else
#define FMUS_PATHSEP '/'
#endif
#endif
#ifndef FMUS_PATHSEP_S
#ifdef FMUS_WINDOZE
#define FMUS_PATHSEP_S "\\"
#else
#define FMUS_PATHSEP_S "/"
#endif
#endif

#ifdef FMUS_WINDOZE
#define FMUS_DOZE_STRERROR_LEN 80
#endif

#ifdef FMUS_DBGMEM2 // all just to suppress a compiler warning
#define FMUS_MALLOCMSG(X,M)           fmus_realloc(NULL, (X), 1, M)
#define FMUS_REALLOCMSG(PTR,SIZE,MSG) fmus_realloc((PTR), (SIZE), 0, (MSG))
#define FMUS_FREEMSG(X,M)             fmus_free((X), M)
#else // ndef FMUS_DBGMEM2
#define FMUS_MALLOCMSG(X,M)           fmus_realloc(NULL, (X), 1)
#define FMUS_REALLOCMSG(PTR,SIZE,MSG) fmus_realloc((PTR), (SIZE), 0)
#define FMUS_FREEMSG(X,M)             fmus_free((X))
#endif

#ifdef FMUS_WINDOZE
#define FMUS_BPRINTF(a,b,...)    gclip_slice_printf((a),NULL,0,(b),__VA_ARGS__) // bug printf
#define FMUS_EPRINTF(a,b,...)    gclip_slice_printf((a),NULL,0,(b),__VA_ARGS__) // error printf
#define FMUS_WPRINTF(a,b,...)    gclip_slice_printf((a),NULL,0,(b),__VA_ARGS__) // warning printf
#define FMUS_MPRINTF(a,b,...)    gclip_slice_printf((a),NULL,0,(b),__VA_ARGS__) // message printf
#define FMUS_XPRINTF(b,...)      printf((b),__VA_ARGS__) // console printf
#else
#define FMUS_BPRINTF(a,c...)     gclip_slice_printf((a),NULL,0,c) // bug printf
#define FMUS_EPRINTF(a,c...)     gclip_slice_printf((a),NULL,0,c) // error printf
#define FMUS_WPRINTF(a,c...)     gclip_slice_printf((a),NULL,0,c) // warning printf
#define FMUS_MPRINTF(a,c...)     gclip_slice_printf((a),NULL,0,c) // message printf
#define FMUS_XPRINTF(c...)       do { printf(c); fflush(stdout); } while (0) // console printf
#endif // ndef FMUS_WINDOZE

typedef s32_t fmus_err_t;

typedef struct fmus_stat {
  u8_t is_dir;
  u8_t is_link;
  off_t fsize;
} fmus_stat_t;

#ifdef FMUS_DBGMEM2
void *fmus_realloc (void *p, size_t size, u8_t zero, const char *label);
void fmus_free (void *p, const char *label);
#else
void *fmus_realloc (void *p, size_t size, u8_t zero);
void fmus_free (void *p);
#endif // ! FMUS_DBGMEM2

#endif
