#ifndef TEXT_H
#define TEXT_H

// libfreetype
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_BITMAP_H
#include FT_STROKER_H

#define TEXT_STYLE_NORMAL      0
#define TEXT_STYLE_OUTLINE     1

#include "fmustypes.h"

typedef struct fmus_font {
  FT_Face *faces;
  u32_t num_faces;
  u8_t *file_contents;
  size_t file_len;
} fmus_font_t;

typedef struct fmus_text {
  FT_Library library;
  fmus_font_t font;
} fmus_text_t;

#include "fmus.h"

fmus_err_t fmus_text_init(fmus_t *fmus, fmus_text_t *text);
fmus_err_t fmus_text_load_font(fmus_t *fmus,
                               fmus_text_t *text,
                               char *fontfile,
                               s32_t desired_face_ix);
void fmus_text_finish(fmus_text_t *text);
fmus_err_t fmus_text_set_pixel_size (fmus_t *fmus, fmus_text_t *text, float size_px, u32_t face_id);
fmus_err_t fmus_text_set_point_size (fmus_t *fmus, fmus_text_t *text, float size_pt, u32_t face_id);
fmus_err_t fmus_text_render (fmus_t *fmus,
                             fmus_text_t *text,
                             s32_t face_ix,
                             u32_t utf32,
                             float px_size,
                             //u8_t print_warnings,
                             s32_t *width_out,
                             s32_t *height_out,
                             u8_t **bitmap_out,
                             s32_t *baseline_y_out,
                             s32_t *advance_out,
                             s32_t *left_bearing_out);
fmus_err_t get_kerning_from_freetype(fmus_t *fmus, fmus_text_t *text, fmus_mpfont_t *mpfont, s32_t font_face_ix);
void set_mpfont_name_from_freetype (fmus_text_t *text, s32_t font_face_ix, char **font_name_inout);

#endif
